require("love_reactor/all")

local pickup_sound = love.audio.newSource("sounds/pickup.wav")
local drop_sound = love.audio.newSource("sounds/drop.wav")
local next_level_sound = love.audio.newSource("sounds/next_level.wav")
local slice_sound = love.audio.newSource("sounds/slick.wav")

function sequence(values)
  local output = {}
  output._values = values
  output._pointer = 1
  output.next = function(self)
    self._pointer = (self._pointer % #self._values) + 1
    return self._values[self._pointer]
  end
  output.previous = function(self)
    self._pointer = self._pointer - 1
    if self._pointer < 1 then
      self._pointer = #self._values
    end
    return self._values[self._pointer]
  end
  output.current = function(self)
    return self._values[self._pointer]
  end
  return output
end

Arrow = class({
}, {drawable, clickable})

backgrounds = sequence({
  love.graphics.newImage("sprites/background.psd"),
  love.graphics.newImage("sprites/background_2.psd"),
  love.graphics.newImage("sprites/background_3.psd"),
})

Background = class({
  sprite = backgrounds:current(),
  offset = {x = 0, y = 0},
  depth = 100000
}, {drawable})

keypressed_stream
  .map(function(e)
    if e == "left" then
      Background.instances.set("sprite", backgrounds:previous())
    elseif e == "right" then
      Background.instances.set("sprite", backgrounds:next())
    end
  end)


Line = class({
  sprite = love.graphics.newImage("sprites/line.psd")
}, {drawable, collidable, draggable, singleton})

LineHelper = class({
  sprite = love.graphics.newImage("sprites/line_helper.psd")
}, {drawable, draggable})

local linearoos = union(Line.instances, LineHelper.instances)
linearoos.aggregate("drag_stream")
  .map(function(e)
    update_stream.take_until(e.member.drag_end)
    .map(function(_)
      linearoos.set("y", love.mouse.getY())
    end)
  end)

Block = class({
  sprite = love.graphics.newImage("sprites/block.psd"),
}, {drawable, collidable, draggable})

collision_stream
  .map(collision_between(Line.instances, Block.instances)).filter(function(e) return e end)
  .map(function(e) return e.b end) -- map to blocks
  .buffer(update_stream) -- blocks collided with this frame
  .buffer_latest(linearoos.aggregate("drag_end")) -- when we drag_end
  .map(function(blocks)
    for _, block in pairs(blocks) do
      block.destroy()
    end

    slice_sound:play()

    if #blocks > 0 then
      for _, block in pairs(Block.instances.values()) do
        if block.position.y < Line.instance.y and not next_mo.contains(block.parent) then
          block.local_position.y = block.local_position.y + Line.instance.height
        end
      end
    end
  end)

linearoos.aggregate("drag_begin")
  .map(function()
    pickup_sound:play()
    collision_stream
      .take_until(linearoos.aggregate("drag_end"))
      .filter(function(e) return e.type == "start" end)
      .map(collision_between(Line.instances, Block.instances))
      .filter(function(e) return e end)
      .map(function(e)
        local cache_color = e.b.color
        e.b.color = {255, 0, 0, 100}
        e.b.collision_stream
          .filter(function(e) return e.type == "end" end)
          .filter(function(e) return Line.instances.contains(e.other) end)
          .combine(linearoos.aggregate("drag_end"))
          .map(function(_)
            e.b.color = cache_color
            return no_more
          end)
      end)
  end)

update_stream
  .map(function()
    for _, tetromino in pairs(Tetromino.instances.values()) do
      for _, block in pairs(tetromino.blocks.values()) do
        block.position = vector.add(tetromino.position, block.local_position)
      end
    end
  end)

Tetromino = class({
  blocks = {},
  init = function(self)
    local blocks = self.blocks
    self.blocks = rxcontainer()
    for _, v in pairs(blocks) do
      self.blocks.add(v)
      v.parent = self
    end
    self.drag_begin = self.blocks.aggregate("drag_begin").map(function(e) return e.value end)
    self.drag_stream = self.blocks.aggregate("drag_stream").map(function(e) return e.value end)
    self.drag_end = self.blocks.aggregate("drag_end").map(function(e) return e.value end)
    self.right_clicks = self.blocks.aggregate("click_stream")
      .filter(function(e) return e.value.type == "down" end)
      .filter(function(e) return e.value.button == "r" end)

    self.right_clicks
      .map(function()
        self:rotate(math.pi / 2)
      end)
  end,

  rotate = function(self, delta)
    local center = table.reduce(table.map(self.blocks.values(), function(block) return block.local_position end), vector.add, {x = 0, y = 0})
    center = vector.mult(center, 1 / #self.blocks.values())
    for _, block in pairs(self.blocks.values()) do
      local d = vector.sub(center, block.local_position)
      local distance = vector.magnitude(d)
      local theta = math.atan2(d.y, d.x)
      theta = theta + delta
      block.local_position.x = center.x + math.cos(theta) * distance
      block.local_position.y = center.y + math.sin(theta) * distance
    end
  end
}, {movable})

Tetromino.instances.aggregate("drag_begin")
  .map(function(e)
    local offset = vector.sub(e.member.position, {x = love.mouse.getX(), y = love.mouse.getY()})
    pickup_sound:play()
    e.member.drag_stream
      .take_until(e.member.drag_end.map(function(e)
        drop_sound:play()
        return "asdf"
      end))
      .map(function()
        e.member.position = vector.add(offset, {x = love.mouse.getX(), y = love.mouse.getY()})
      end)
  end)

next_mo = rxcontainer()

next_mo.aggregate("drag_begin")
  .map(function(e)
    next_mo.remove(e.member)
    local new = make_tetromino(table.pick_random(tetrominos))
    new.position = e.member.position
    next_mo.add(new)
  end)

tetrominos = {
  line = {
    tile_width = 32,
    tile_height = 32,
    data = {
      "####"
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.blue
      }
    },
  },
  tee = {
    tile_width = 32,
    tile_height = 32,
    data = {
      "###",
      " # "
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.yellow
      }
    },
  },
  ess = {
    tile_width = 32,
    tile_height = 32,
    data = {
      " ##",
      "## "
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.red
      }
    },
  },
  square = {
    tile_width = 32,
    tile_height = 32,
    data = {
      "##",
      "##"
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.purple
      }
    },
  },
  gamma = {
    tile_width = 32,
    tile_height = 32,
    data = {
      "##",
      "# ",
      "# "
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.orange
      }
    },
  },
  ell = {
    tile_width = 32,
    tile_height = 32,
    data = {
      "# ",
      "# ",
      "##"
    },
    legend = {
      ["#"] = {
        class = Block,
        color = colors.white
      }
    },
  }
}

tetrominos = table.map(tetrominos, parse_map)


function make_tetromino(map_data)
  local depth = -#Tetromino.instances.values()
  local blocks = simple_instantiate(map_data)

  for _, block in pairs(blocks) do
    block.local_position = block.position
    block.depth = depth
  end

  local output = Tetromino.new({
    width = map_data.width,
    height = map_data.height,
    depth = depth,
    blocks = blocks
  })

  return output
end

function love.load()
  Background.new()
  love.window.setMode(600, 800)
  local t_mo = make_tetromino(tetrominos.line)
  t_mo.x = 400
  t_mo.y = 75
  next_mo.add(t_mo)
  local line = Line.new({
    x = love.window.getWidth() / 2,
    y = love.window.getHeight() - 150
  })

  LineHelper.new({
    x = love.window.getWidth() - 50,
    y = line.y
  })

  local left_arrow = Arrow.new({
    sprite = love.graphics.newImage("sprites/left_arrow.psd"),
    x = 50,
    y = love.window.getHeight() - 50
  })
  left_arrow.click_stream
    .filter(function(e) return e.type == "down" end)
    .map(function()
      next_level_sound:play()
      Background.instances.set("sprite", backgrounds:previous())
    end)

  local right_arrow = Arrow.new({
    sprite = love.graphics.newImage("sprites/right_arrow.psd"),
    x = love.window.getWidth() - 50,
    y = love.window.getHeight() - 50
  })
  right_arrow.click_stream
    .filter(function(e) return e.type == "down" end)
    .map(function()
      next_level_sound:play()
      Background.instances.set("sprite", backgrounds:next())
    end)
end
